@@ -12,6 +12,8 @@
 import logging
 import re
 from abc import ABCMeta, abstractmethod
+from dataclasses import dataclass
+from datetime import datetime
 from enum import Enum
 
 from common.utils import time_is_up
@@ -36,32 +38,6 @@ EC2_SCHEDULED_EVENT_CODES = [
 CONFIG_FILE_DIR = "/etc/parallelcluster/slurm_plugin"
 
 
-class EC2Instance:
-    def __init__(self, id, private_ip, hostname, launch_time):
-        """Initialize slurm node with attributes."""
-        self.id = id
-        self.private_ip = private_ip
-        self.hostname = hostname
-        self.launch_time = launch_time
-        self.slurm_node = None
-
-    def __eq__(self, other):
-        """Compare 2 SlurmNode objects."""
-        if isinstance(other, EC2Instance):
-            return self.__dict__ == other.__dict__
-        return False
-
-    def __repr__(self):
-        attrs = ", ".join(["{key}={value}".format(key=key, value=repr(value)) for key, value in self.__dict__.items()])
-        return "{class_name}({attrs})".format(class_name=self.__class__.__name__, attrs=attrs)
-
-    def __str__(self):
-        return f"{self.id}"
-
-    def __hash__(self):
-        return hash(self.id)
-
-
 class PartitionStatus(Enum):
     UP = "UP"
     DOWN = "DOWN"
@@ -94,8 +70,8 @@ class SlurmPartition:
                     node.is_healthy(terminate_drain_nodes, terminate_down_nodes, log_warn_if_unhealthy=False)
                     and node.is_online()
                 ):
-                    logger.debug("Currently online node: %s, node state: %s", node.name, node.state)
-                    online_compute_resources.add(node.get_instance_name())
+                    logger.debug("Currently online node: %s, node state: %s", node.name, node.state_string)
+                    online_compute_resources.add(node.compute_resource_name)
         return online_compute_resources
 
     def __eq__(self, other):
@@ -106,29 +82,48 @@ class SlurmPartition:
 
 
 class SlurmNode(metaclass=ABCMeta):
-    SLURM_SCONTROL_BUSY_STATES = {"MIXED", "ALLOCATED", "COMPLETING"}
+    SLURM_SCONTROL_COMPLETING_STATE = "COMPLETING"
+    SLURM_SCONTROL_BUSY_STATES = {"MIXED", "ALLOCATED", SLURM_SCONTROL_COMPLETING_STATE}
     SLURM_SCONTROL_IDLE_STATE = "IDLE"
     SLURM_SCONTROL_DOWN_STATE = "DOWN"
     SLURM_SCONTROL_DRAIN_STATE = "DRAIN"
     SLURM_SCONTROL_POWERING_DOWN_STATE = "POWERING_DOWN"
-    SLURM_SCONTROL_POWER_STATE = "IDLE+CLOUD+POWER"
-    SLURM_SCONTROL_POWER_UP_STATE = "#"
+    SLURM_SCONTROL_POWER_DOWN_STATE = "POWER_DOWN"
+    SLURM_SCONTROL_POWERED_DOWN_STATE = "POWERED_DOWN"
+    SLURM_SCONTROL_POWER_UP_STATE = "POWERING_UP"
     SLURM_SCONTROL_ONLINE_STATES = {"IDLE+CLOUD", "MIXED+CLOUD", "ALLOCATED+CLOUD", "COMPLETING+CLOUD"}
-    SLURM_SCONTROL_POWER_WITH_JOB_STATE = "MIXED+CLOUD+POWER"
-    SLURM_SCONTROL_RESUME_FAILED_STATE = "DOWN*+CLOUD+POWER"
+    SLURM_SCONTROL_POWER_WITH_JOB_STATE = {"MIXED", "CLOUD", "POWERED_DOWN"}
+    SLURM_SCONTROL_RESUME_FAILED_STATE = {"DOWN", "CLOUD", "POWERED_DOWN", "NOT_RESPONDING"}
+    # Due to a bug in Slurm a powered down node can enter IDLE+CLOUD+POWER_DOWN+POWERED_DOWN state
+    SLURM_SCONTROL_POWER_STATES = [{"IDLE", "CLOUD", "POWERED_DOWN"}, {"IDLE", "CLOUD", "POWERED_DOWN", "POWER_DOWN"}]
+    SLURM_SCONTROL_REBOOT_REQUESTED_STATE = "REBOOT_REQUESTED"
+    SLURM_SCONTROL_REBOOT_ISSUED_STATE = "REBOOT_ISSUED"
 
-    def __init__(self, name, nodeaddr, nodehostname, state, partitions=None, instance=None):
+    EC2_ICE_ERROR_CODES = {
+        "InsufficientInstanceCapacity",
+        "InsufficientHostCapacity",
+        "InsufficientReservedInstanceCapacity",
+        "MaxSpotInstanceCountExceeded",
+        "Unsupported",
+        "SpotMaxPriceTooLow",
+    }
+
+    def __init__(self, name, nodeaddr, nodehostname, state, partitions=None, reason=None, instance=None):
         """Initialize slurm node with attributes."""
         self.name = name
         self.nodeaddr = nodeaddr
         self.nodehostname = nodehostname
-        self.state = state
+        self.state_string = state
+        self.states = set(state.split("+"))
         self.partitions = partitions.strip().split(",") if partitions else None
+        self.reason = reason
         self.instance = instance
         self.is_static_nodes_in_replacement = False
-        self._is_being_replaced = False
+        self.is_being_replaced = False
         self._is_replacement_timeout = False
         self.is_failing_health_check = False
+        self.error_code = self._parse_error_code()
+        self.queue_name, self._node_type, self.compute_resource_name = parse_nodename(name)
 
     def is_nodeaddr_set(self):
         """Check if nodeaddr(private ip) for the node is set."""
@@ -136,11 +131,11 @@ class SlurmNode(metaclass=ABCMeta):
 
     def has_job(self):
         """Check if slurm node is in a working state."""
-        return any(working_state in self.state for working_state in self.SLURM_SCONTROL_BUSY_STATES)
+        return any(working_state in self.states for working_state in self.SLURM_SCONTROL_BUSY_STATES)
 
     def _is_drain(self):
         """Check if slurm node is in any drain(draining, drained) states."""
-        return self.SLURM_SCONTROL_DRAIN_STATE in self.state
+        return self.SLURM_SCONTROL_DRAIN_STATE in self.states
 
     def is_drained(self):
         """
@@ -148,21 +143,39 @@ class SlurmNode(metaclass=ABCMeta):
 
         drained(sinfo) is equivalent to IDLE+DRAIN(scontrol) or DOWN+DRAIN(scontrol)
         """
-        return self._is_drain() and (
-            self.SLURM_SCONTROL_IDLE_STATE in self.state or self.SLURM_SCONTROL_DOWN_STATE in self.state
+        return (
+            self._is_drain()
+            and (self.SLURM_SCONTROL_IDLE_STATE in self.states or self.is_down())
+            and not self.is_completing()
         )
 
+    def is_completing(self):
+        """Check if slurm node is in COMPLETING state."""
+        return self.SLURM_SCONTROL_COMPLETING_STATE in self.states
+
+    def is_power_down(self):
+        """Check if slurm node is in power down state."""
+        return self.SLURM_SCONTROL_POWER_DOWN_STATE in self.states
+
     def is_powering_down(self):
         """Check if slurm node is in powering down state."""
-        return self.SLURM_SCONTROL_POWERING_DOWN_STATE in self.state
+        return self.SLURM_SCONTROL_POWERING_DOWN_STATE in self.states
+
+    def is_powered_down(self):
+        """Check if slurm node is in powered down state."""
+        return self.SLURM_SCONTROL_POWERED_DOWN_STATE in self.states
 
     def is_power(self):
         """Check if slurm node is in power state."""
-        return self.SLURM_SCONTROL_POWER_STATE == self.state
+        return self.states in self.SLURM_SCONTROL_POWER_STATES
 
     def is_down(self):
         """Check if slurm node is in a down state."""
-        return self.SLURM_SCONTROL_DOWN_STATE in self.state and not self.is_powering_down()
+        return (
+            self.SLURM_SCONTROL_DOWN_STATE in self.states
+            and not self.is_powering_down()
+            and (not self.is_power_down() or self.is_powered_down())
+        )
 
     def is_up(self):
         """Check if slurm node is in a healthy state."""
@@ -170,11 +183,11 @@ class SlurmNode(metaclass=ABCMeta):
 
     def is_powering_up(self):
         """Check if slurm node is in powering up state."""
-        return self.SLURM_SCONTROL_POWER_UP_STATE in self.state
+        return self.SLURM_SCONTROL_POWER_UP_STATE in self.states
 
     def is_online(self):
         """Check if slurm node is online with backing instance."""
-        return self.state in self.SLURM_SCONTROL_ONLINE_STATES
+        return self.state_string in self.SLURM_SCONTROL_ONLINE_STATES
 
     def is_configuring_job(self):
         """Check if slurm node is configuring with job and haven't begun to run a job."""
@@ -182,7 +195,7 @@ class SlurmNode(metaclass=ABCMeta):
 
     def is_power_with_job(self):
         """Dynamic nodes allocated a job but power up process has not started yet."""
-        return self.state == self.SLURM_SCONTROL_POWER_WITH_JOB_STATE
+        return self.states == self.SLURM_SCONTROL_POWER_WITH_JOB_STATE
 
     def is_running_job(self):
         """Check if slurm node is running a job but not in configuring job state."""
@@ -190,7 +203,36 @@ class SlurmNode(metaclass=ABCMeta):
 
     def is_resume_failed(self):
         """Check if node resume timeout expires."""
-        return self.state == self.SLURM_SCONTROL_RESUME_FAILED_STATE
+        return self.states == self.SLURM_SCONTROL_RESUME_FAILED_STATE
+
+    def is_poweing_up_idle(self):
+        """Check if node is in IDLE# state."""
+        return self.SLURM_SCONTROL_IDLE_STATE in self.states and self.is_powering_up()
+
+    def is_ice(self):
+        return self.error_code in self.EC2_ICE_ERROR_CODES
+
+    def is_reboot_requested(self):
+        return self.SLURM_SCONTROL_REBOOT_REQUESTED_STATE in self.states
+
+    def is_reboot_issued(self):
+        return self.SLURM_SCONTROL_REBOOT_ISSUED_STATE in self.states
+
+    def is_rebooting(self):
+        """
+        Check if the node is rebooting via scontrol reboot.
+
+        Check that the node is in a state consistent with the scontrol reboot request.
+        """
+        cond = False
+        if self.is_reboot_issued() or self.is_reboot_requested():
+            logger.debug(
+                "Node state check: node %s is currently rebooting, ignoring, node state: %s",
+                self,
+                self.state_string,
+            )
+            cond = True
+        return cond
 
     @abstractmethod
     def is_state_healthy(self, terminate_drain_nodes, terminate_down_nodes, log_warn_if_unhealthy=True):
@@ -213,10 +255,11 @@ class SlurmNode(metaclass=ABCMeta):
         """Check if a slurm node is considered healthy."""
         pass
 
-    @abstractmethod
     def is_powering_down_with_nodeaddr(self):
         """Check if a slurm node is a powering down node with instance backing."""
-        pass
+        # Node in POWERED_DOWN with nodeaddr still set, may have not been seen during the POWERING_DOWN transition
+        # for example because of a short SuspendTimeout
+        return self.is_nodeaddr_set() and (self.is_power() or self.is_powering_down())
 
     def is_backing_instance_valid(self, log_warn_if_unhealthy=True):
         """Check if a slurm node's addr is set, it points to a valid instance in EC2."""
@@ -226,7 +269,7 @@ class SlurmNode(metaclass=ABCMeta):
                     logger.warning(
                         "Node state check: no corresponding instance in EC2 for node %s, node state: %s",
                         self,
-                        self.state,
+                        self.state_string,
                     )
                 return False
         return True
@@ -236,10 +279,13 @@ class SlurmNode(metaclass=ABCMeta):
         """Check if the node need to be reset if node is inactive."""
         pass
 
-    def get_instance_name(self):
-        """Get instance name of given node."""
-        _, _, instance_name = parse_nodename(self.name)
-        return instance_name
+    def _parse_error_code(self):
+        """Parse RunInstance error code from node reason."""
+        if self.reason and self.reason.startswith("(Code:"):
+            index_of_bracket = self.reason.find(")")
+            error_code = self.reason[len("(Code:") : index_of_bracket]  # noqa E203: whitespace before ':'
+            return error_code
+        return None
 
     def __eq__(self, other):
         """Compare 2 SlurmNode objects."""
@@ -259,9 +305,9 @@ class SlurmNode(metaclass=ABCMeta):
 
 
 class StaticNode(SlurmNode):
-    def __init__(self, name, nodeaddr, nodehostname, state, partitions=None, instance=None):
+    def __init__(self, name, nodeaddr, nodehostname, state, partitions=None, reason=None, instance=None):
         """Initialize slurm node with attributes."""
-        super().__init__(name, nodeaddr, nodehostname, state, partitions, instance)
+        super().__init__(name, nodeaddr, nodehostname, state, partitions, reason, instance)
 
     def is_healthy(self, terminate_drain_nodes, terminate_down_nodes, log_warn_if_unhealthy=True):
         """Check if a slurm node is considered healthy."""
@@ -275,31 +321,34 @@ class StaticNode(SlurmNode):
 
     def is_state_healthy(self, terminate_drain_nodes, terminate_down_nodes, log_warn_if_unhealthy=True):
         """Check if a slurm node's scheduler state is considered healthy."""
-        # Check to see if node is in DRAINED, ignoring any node currently being replaced
-        if self.is_drained() and terminate_drain_nodes:
-            if self._is_being_replaced:
+        # Check if node is rebooting: if so, the node is healthy
+        if self.is_rebooting():
+            return True
+        # Check to see if node is in DRAINED, ignoring any node currently being replaced or in POWER_DOWN
+        if self.is_drained() and not self.is_power_down() and terminate_drain_nodes:
+            if self.is_being_replaced:
                 logger.debug(
                     "Node state check: node %s in DRAINED but is currently being replaced, ignoring, node state: %s",
                     self,
-                    self.state,
+                    self.state_string,
                 )
                 return True
             else:
                 if log_warn_if_unhealthy:
-                    logger.warning("Node state check: node %s in DRAINED, node state: %s", self, self.state)
+                    logger.warning("Node state check: node %s in DRAINED, node state: %s", self, self.state_string)
                 return False
         # Check to see if node is in DOWN, ignoring any node currently being replaced
         elif self.is_down() and terminate_down_nodes:
-            if self._is_being_replaced:
+            if self.is_being_replaced:
                 logger.debug(
                     "Node state check: node %s in DOWN but is currently being replaced, ignoring. Node state: ",
                     self,
-                    self.state,
+                    self.state_string,
                 )
                 return True
             else:
                 if log_warn_if_unhealthy:
-                    logger.warning("Node state check: node %s in DOWN, node state: %s", self, self.state)
+                    logger.warning("Node state check: node %s in DOWN, node state: %s", self, self.state_string)
                 return False
         return True
 
@@ -308,7 +357,9 @@ class StaticNode(SlurmNode):
         if not self.is_nodeaddr_set():
             if log_warn_if_unhealthy:
                 logger.warning(
-                    "Node state check: static node without nodeaddr set, node %s, node state %s:", self, self.state
+                    "Node state check: static node without nodeaddr set, node %s, node state %s:",
+                    self,
+                    self.state_string,
                 )
             return False
         return True
@@ -320,7 +371,7 @@ class StaticNode(SlurmNode):
             logger.warning(
                 "Node bootstrap error: Node %s is currently in replacement and no backing instance, node state %s:",
                 self,
-                self.state,
+                self.state_string,
             )
             return True
             # Replacement timeout expires for node in replacement
@@ -328,47 +379,46 @@ class StaticNode(SlurmNode):
             logger.warning(
                 "Node bootstrap error: Replacement timeout expires for node %s in replacement, node state %s:",
                 self,
-                self.state,
+                self.state_string,
             )
             return True
         elif self.is_failing_health_check and self.is_static_nodes_in_replacement:
             logger.warning(
                 "Node bootstrap error: Node %s failed during bootstrap when performing health check, node state %s:",
                 self,
-                self.state,
+                self.state_string,
             )
             return True
         return False
 
-    def is_powering_down_with_nodeaddr(self):
-        """Check if a slurm node is a powering down node with instance backing. Static node will not powering down."""
-        return False
-
     def needs_reset_when_inactive(self):
         """Check if the node need to be reset if node is inactive."""
         return self.is_nodeaddr_set()
 
 
 class DynamicNode(SlurmNode):
-    def __init__(self, name, nodeaddr, nodehostname, state, partitions=None, instance=None):
+    def __init__(self, name, nodeaddr, nodehostname, state, partitions=None, reason=None, instance=None):
         """Initialize slurm node with attributes."""
-        super().__init__(name, nodeaddr, nodehostname, state, partitions, instance)
+        super().__init__(name, nodeaddr, nodehostname, state, partitions, reason, instance)
 
     def is_state_healthy(self, terminate_drain_nodes, terminate_down_nodes, log_warn_if_unhealthy=True):
         """Check if a slurm node's scheduler state is considered healthy."""
-        # Check to see if node is in DRAINED, ignoring any node currently being replaced
-        if self.is_drained() and terminate_drain_nodes:
+        # Check if node is rebooting: if so, the node is healthy
+        if self.is_rebooting():
+            return True
+        # Check to see if node is in DRAINED, ignoring any node currently being replaced or in POWER_DOWN
+        if self.is_drained() and not self.is_power_down() and terminate_drain_nodes:
             if log_warn_if_unhealthy:
-                logger.warning("Node state check: node %s in DRAINED, node state: %s", self, self.state)
+                logger.warning("Node state check: node %s in DRAINED, node state: %s", self, self.state_string)
             return False
         # Check to see if node is in DOWN, ignoring any node currently being replaced
         elif self.is_down() and terminate_down_nodes:
             if not self.is_nodeaddr_set():
                 # Silently handle failed to launch dynamic node to clean up normal logging
-                logger.debug("Node state check: node %s in DOWN, node state: %s", self, self.state)
+                logger.debug("Node state check: node %s in DOWN, node state: %s", self, self.state_string)
             else:
                 if log_warn_if_unhealthy:
-                    logger.warning("Node state check: node %s in DOWN, node state: %s", self, self.state)
+                    logger.warning("Node state check: node %s in DOWN, node state: %s", self, self.state_string)
             return False
         return True
 
@@ -381,32 +431,32 @@ class DynamicNode(SlurmNode):
     def is_bootstrap_failure(self):
         """Check if a slurm node has boostrap failure."""
         # no backing instance + [working state]# in node state
-        if self.is_configuring_job() and not self.is_backing_instance_valid(log_warn_if_unhealthy=False):
+        if (self.is_configuring_job() or self.is_poweing_up_idle()) and not self.is_backing_instance_valid(
+            log_warn_if_unhealthy=False
+        ):
             logger.warning(
                 "Node bootstrap error: Node %s is in power up state without valid backing instance, node state: %s",
                 self,
-                self.state,
+                self.state_string,
             )
             return True
-        # Dynamic node in DOWN*+CLOUD+POWER state
+        # Dynamic node in DOWN+CLOUD+POWERED_DOWN+NOT_RESPONDING state
         elif self.is_resume_failed() and self.is_nodeaddr_set():
             # We need to check if nodeaddr is set to avoid counting powering up nodes as bootstrap failure nodes during
             # cluster start/stop.
-            logger.warning("Node bootstrap error: Resume timeout expires for node %s, node state: %s", self, self.state)
+            logger.warning(
+                "Node bootstrap error: Resume timeout expires for node %s, node state: %s", self, self.state_string
+            )
             return True
         elif self.is_failing_health_check and self.is_powering_up():
             logger.warning(
                 "Node bootstrap error: Node %s failed during bootstrap when performing health check, node state: %s",
                 self,
-                self.state,
+                self.state_string,
             )
             return True
         return False
 
-    def is_powering_down_with_nodeaddr(self):
-        """Check if a slurm node is a powering down node with instance backing."""
-        return self.is_nodeaddr_set() and (self.is_power() or self.is_powering_down())
-
     def needs_reset_when_inactive(self):
         """Check if the node need to be reset if node is inactive."""
         return self.is_nodeaddr_set() or (not (self.is_power() or self.is_powering_down() or self.is_down()))
@@ -459,19 +509,25 @@ class InvalidNodenameError(ValueError):
     r"""
     Exception raised when encountering a NodeName that is invalid/incorrectly formatted.
 
-    Valid NodeName format: {queue-name}-{st/dy}-{instancetype}-{number}
-    And match: ^([a-z0-9\-]+)-(st|dy)-([a-z0-9-]+)-\d+$
-    Sample NodeName: queue-1-st-c5xlarge-2
+    Valid NodeName format: {queue-name}-{st/dy}-{compute-resource}-{number}
+    And match: ^([a-z0-9\-]+)-(st|dy)-([a-z0-9\-]+)-\d+$
+    Sample NodeName: queue-1-st-computeresource-2
     """
 
     pass
 
 
+@dataclass
+class ComputeResourceFailureEvent:
+    timestamp: datetime
+    error_code: str
+
+
 def parse_nodename(nodename):
     """Parse queue_name, node_type (st vs dy) and instance_type from nodename."""
-    nodename_capture = re.match(r"^([a-z0-9\-]+)-(st|dy)-([a-z0-9]+)-\d+$", nodename)
+    nodename_capture = re.match(r"^([a-z0-9\-]+)-(st|dy)-([a-z0-9\-]+)-\d+$", nodename)
     if not nodename_capture:
         raise InvalidNodenameError
 
-    queue_name, node_type, instance_name = nodename_capture.groups()
-    return queue_name, node_type, instance_name
+    queue_name, node_type, compute_resource_name = nodename_capture.groups()
+    return queue_name, node_type, compute_resource_name
